//---------------------------------------------------------------------------
#include <stdio.h>
#include <string.h>
#include <stdlib.h>

#include "common.h"
//---------------------------------------------------------------------------
int parseNumber(char *str);
int readFile(char *fileName, BYTE *buff, int minLen, int maxLen);
int writeFile(char *fileName, BYTE *buff, int len, BOOL append);
int printError(int err, void *par);
void version(void);
void usage(void);

//---------------------------------------------------------------------------
int main(int argc, char* argv[])
{
  char *fnIn = NULL, *fnOut = NULL, fn[256], n[9], *pc;
  BOOL vb = FALSE, append = FALSE, ptp = FALSE;
  BYTE fileBuff[MAX_FILE_LENGTH], buffIn[MAX_FILE_LENGTH];
  BYTE buffOut[MAX_FILE_LENGTH + 128], *pb;
  int i, err, mode, lenIn, lenOut[4], lenRt[4], lenTot;
  int unpackRtnAddr = -1, firstLine = 0, lastLine = 255, lineAdr;
  int useModes = (PACK_MODE_1 | PACK_MODE_2 | PACK_MODE_3 | PACK_MODE_4);

  // spracovanie a kontrola vstupnych parametrov
  err = 0;
  do {
    for (i = 1; i < argc; i++) {
      if (strcmp(argv[i], "-h") == 0) {
        err = ERR_PRINT_USAGE;
        break;
      }
      else if (strcmp(argv[i], "-v") == 0) {
        err = ERR_PRINT_VERSION;
        break;
      }
      else if (strcmp(argv[i], "-vb") == 0) {
        vb = TRUE;
        continue;
      }
      else if (strcmp(argv[i], "-append") == 0) {
        append = TRUE;
        continue;
      }
      else if (strcmp(argv[i], "-ptp") == 0) {
        ptp = TRUE;
        continue;
      }
      else if (strcmp(argv[i], "-m") == 0) {
        useModes = 0;
        if ((i + 1) < argc) {
          pc = argv[++i];
          while (*pc) {
            if (*pc < '1' || *pc > '4')
              break;
            if (useModes & (1 << (*pc - '1')))
              break;
            useModes |= (1 << (*pc - '1'));
            pc++;
          }
        }
        if (useModes == 0 || *pc != 0) {
          err = printError(ERR_BAD_PACK_MODE, argv[i]);
          break;
        }
        continue;
      }
      else if (strcmp(argv[i], "-a") == 0) {
        if (++i == argc)
          unpackRtnAddr = -1;
        else
          unpackRtnAddr = parseNumber(argv[i]);
        if (unpackRtnAddr < 0 || unpackRtnAddr > 65535) {
          err = printError(ERR_BAD_ADDRESS, argv[i - 1]);
          break;
        }
      }
      else if (strcmp(argv[i], "-lf") == 0) {
        if (++i == argc)
          firstLine = -1;
        else
          firstLine = parseNumber(argv[i]);
        if (firstLine < 0 || firstLine > 255) {
          err = printError(ERR_BAD_LINE_VALUE, argv[i - 1]);
          break;
        }
      }
      else if (strcmp(argv[i], "-ll") == 0) {
        if (++i == argc)
          lastLine = -1;
        else
          lastLine = parseNumber(argv[i]);
        if (lastLine < 0 || lastLine > 255) {
          err = printError(ERR_BAD_LINE_VALUE, argv[i - 1]);
          break;
        }
      }
      else if (strcmp(argv[i], "-o") == 0) {
        if (++i == argc || fnOut != NULL) {
          err = printError(ERR_OUTPUT_FILE_MORE, NULL);
          break;
        }
        fnOut = argv[i];
        continue;
      }
      else {
        if (fnIn != NULL) {
          err = printError(ERR_INPUT_FILE_MORE, NULL);
          break;
        }
        fnIn = argv[i];
        continue;
      }
    }
    if (err != 0)
      break;

    if (fnIn == NULL) {
      err = printError(ERR_NO_INPUT_FILE, NULL);
      break;
    }

    if (firstLine > lastLine) {
      err = printError(ERR_BAD_LINES_RANGE, NULL);
      break;
    }

    if (fnOut == NULL) {
      strcpy(fn, fnIn);
      pc = strrchr(fn, '.');
      if (pc != NULL)
        *pc = 0;
      if (ptp)
        strcat(fn, ".ptp");
      else
        strcat(fn, ".pck");
      fnOut = fn;
      if (strcmp(fnOut, fnIn) == 0)
        strcat(fnOut, ".pck");
    }

    if (unpackRtnAddr >= 0)
      getRoutinesLength(lenRt);
    else
      memset(lenRt, 0, sizeof(lenRt));

    // precitanie vstupneho suboru
    if (vb)
      printf("Reading input file '%s'\n", fnIn);
    lenIn = readFile(fnIn, fileBuff, MIN_FILE_LENGTH, MAX_FILE_LENGTH);
    if (lenIn < 0) {
      err = lenIn;
      break;
    }
    lenIn = (lastLine - firstLine + 1) * 48; // skutocna dlzka obrazovych dat

    if (vb)
      printf("Source graphics data length: %d bytes\n", lenIn);

    // spakovanie dat zvolenymi modmi
    for (i = 0; i < 4; i++) {
      mode = useModes & (1 << i);
      if (mode) {
        if (vb)
          printf("Packing in mode: %d\n", i + 1);
        memcpy(buffIn, fileBuff, sizeof(buffIn));
        lenOut[i] = pack(buffOut, buffIn, firstLine, lastLine, mode);

        if (vb) {
          if (lenOut[i] < 0)
            printf("  Packing ineffective!\n");
          else {
            printf("  Packed data length: %d bytes", lenOut[i]);
            if (unpackRtnAddr >= 0) {
              printf("\n  Unpack routine length: %d bytes", lenRt[i]);
              lenTot = lenRt[i] + lenOut[i];
              printf("\n  Total length: %d bytes", lenTot);
            }
            else
              lenTot = lenOut[i];
            printf(" (%.2f%%)\n", (lenTot * 100.0) / lenIn);
            lenOut[i] = lenTot;
          }
        }
        else if (lenOut[i] > 0 && unpackRtnAddr >= 0) {
          lenTot = lenRt[i] + lenOut[i];
          lenOut[i] = lenTot;
        }
      }
      else
        lenOut[i] = -1;
    }

    // najdenie najlepsieho vysledku
    lenTot = lenIn;
    mode = -1;
    for (i = 0; i < 4; i++) {
      if (lenOut[i] > 0 && lenOut[i] < lenTot) {
        lenTot = lenOut[i];
        mode = i;
      }
    }
    if (mode < 0) {
      err = printError(ERR_INEFFECTIVE, NULL);
      break;
    }

    // konecne pakovanie najlepsim modom
    memcpy(buffIn, fileBuff, sizeof(buffIn));
    lenOut[0] = pack(buffOut, buffIn, firstLine, lastLine, 1 << mode);
    lenTot = lenOut[0];
    if (unpackRtnAddr >= 0)
      lenTot += lenRt[mode];
    printf("Packed in mode: %d\n", mode + 1);
    printf("Original graphics data length: %d bytes\n", lenIn);
    printf("New length: %d bytes (%.2f%%)\n", lenTot, (lenTot * 100.0) / lenIn);

    if (mode < 2)
      lineAdr = 0xC000 + firstLine * 64;
    else
      lineAdr = 0xC000 + lastLine * 64 + 47;

    if (ptp) {
      if (vb)
        printf("Constructing PTP file\n");

      pc = strrchr(fnIn, '/');
      if (pc == NULL)
        pc = strrchr(fnIn, '\\');
      if (pc == NULL)
        pc = fnIn;
      else
        pc++;
      strncpy(n, pc, 8);
      n[8] = 0;
      pc = strrchr(n, '.');
      if (pc != NULL)
        *pc = 0;

      // PMD 85 PTP
      memcpy(buffIn, buffOut, lenOut[0]);
      pb = buffOut;
      constructCode(&pb, buffIn, lenOut[0],
                    (unpackRtnAddr >= 0) ? (1 << mode) : 0,
                    unpackRtnAddr, lineAdr, n);
      lenTot = pb - buffOut;
    }
    else if (unpackRtnAddr >= 0) {
      memcpy(buffIn, buffOut, lenOut[0]);
      pb = buffOut;
      constructBin(&pb, buffIn, lenOut[0], (1 << mode), unpackRtnAddr, lineAdr);
      lenTot = pb - buffOut;
    }

    // zapis vystupneho suboru
    if (vb)
      printf("Writing to the output file '%s'\n", fnOut);
    err = writeFile(fnOut, buffOut, lenTot, append);

  } while (FALSE);

  if (err == ERR_PRINT_USAGE) {
    usage();
    err = 0;
  }
  if (err == ERR_PRINT_VERSION) {
    version();
    err = 0;
  }

//  getchar();

  return -err;
}
//---------------------------------------------------------------------------
int parseNumber(char *str)
{
  char *p;
  int val = strtol(str, &p, 0);
  if (val < 0 || (val == 0 && *p != 0))
    val = -1;
  return val;
}
//---------------------------------------------------------------------------
int readFile(char *fileName, BYTE *buff, int minLen, int maxLen)
{
  int err = 0, len;
  FILE *f;

  do {
    f = fopen(fileName, "rb");
    if (f == NULL) {
      err = printError(ERR_OPEN_SRC_FILE, fileName);
      break;
    }
    if (fseek(f, 0L, SEEK_END) != 0) {
      err = printError(ERR_ACCESS_SRC_FILE, fileName);
      break;
    }
    len = ftell(f);
    if (len < 0) {
      err = printError(ERR_LENGTH_SRC_FILE, fileName);
      break;
    }
    if (len < minLen || len > maxLen) {
      err = printError(ERR_BAD_LEN_SRC_FILE, fileName);
      break;
    }
    if (fseek(f, 0L, SEEK_SET) != 0) {
      err = printError(ERR_ACCESS_SRC_FILE, fileName);
      break;
    }
    if (fread(buff, 1, len, f) != (unsigned int)len) {
      err = printError(ERR_READ_SRC_FILE, fileName);
      break;
    }
  } while (FALSE);

  fclose(f);

  if  (err < 0)
    return err;

  return len;
}
//---------------------------------------------------------------------------
int writeFile(char *fileName, BYTE *buff, int len, BOOL append)
{
  int err = 0;
  FILE *f;

  do {
    if (append)
      f = fopen(fileName, "ab");
    else
      f = fopen(fileName, "wb");
    if (f == NULL) {
      err = printError(ERR_OPEN_DEST_FILE, fileName);
      break;
    }
    if (!append && fseek(f, 0L, SEEK_SET) != 0) {
      err = printError(ERR_ACCESS_DEST_FILE, fileName);
      break;
    }
    if (fwrite(buff, 1, len, f) != (unsigned int)len) {
      err = printError(ERR_WRITE_DEST_FILE, fileName);
      break;
    }
  } while (FALSE);

  fclose(f);

  return err;
}
//---------------------------------------------------------------------------
int printError(int err, void *par)
{
  switch (err) {
    case ERR_BAD_PACK_MODE :
      fprintf(stderr, "Bad packing mode specified: '%s'!", (char *)par);
      break;

    case ERR_BAD_OPTIONS :
      fprintf(stderr, "Bad combination of an options!");
      break;

    case ERR_OPEN_SRC_FILE :
      fprintf(stderr, "Can't open input file '%s'!", (char *)par);
      break;

    case ERR_ACCESS_SRC_FILE :
      fprintf(stderr, "Can't access input file '%s'!", (char *)par);
      break;

    case ERR_LENGTH_SRC_FILE :
      fprintf(stderr, "Can't retrieve length of the input file '%s'!", (char *)par);
      break;

    case ERR_BAD_LEN_SRC_FILE :
      fprintf(stderr, "Length of the input file '%s' must be 16384!", (char *)par);
      break;

    case ERR_READ_SRC_FILE :
      fprintf(stderr, "Can't read from the input file '%s'!", (char *)par);
      break;

    case ERR_OPEN_DEST_FILE :
      fprintf(stderr, "Can't open output file '%s'!", (char *)par);
      break;

    case ERR_ACCESS_DEST_FILE :
      fprintf(stderr, "Can't access output file '%s'!", (char *)par);
      break;

    case ERR_WRITE_DEST_FILE :
      fprintf(stderr, "Can't write to the output file '%s'!", (char *)par);
      break;

    case ERR_BAD_ADDRESS :
      fprintf(stderr, "After '%s' must follow address in range 0 to 65535!", (char *)par);
      break;

    case ERR_BAD_LINE_VALUE :
      fprintf(stderr, "After '%s' must follow microline in range 0 to 255!", (char *)par);
      break;

    case ERR_BAD_LINES_RANGE :
      fprintf(stderr, "First line number must be lower or equal to last line number!");
      break;

    case ERR_INEFFECTIVE :
      fprintf(stderr, "Packing ineffective!");
      break;

    case ERR_NO_INPUT_FILE :
      fprintf(stderr, "Input file not specified!\nTry 'quido -h' for help.");
      break;

    case ERR_INPUT_FILE_MORE :
      fprintf(stderr, "Only one input file can be specified!");
      break;

    case ERR_OUTPUT_FILE_MORE :
      fprintf(stderr, "Only one output file can be specified!");
      break;
  }

  fprintf(stderr, "\n\n");

  return err;
}
//---------------------------------------------------------------------------
void version(void)
{
	printf("%s v%d.%d - %s\n",
         PROG_NAME, PROG_VERSION / 16, PROG_VERSION % 16, PROG_AUTHOR);
}
//---------------------------------------------------------------------------
void pressEnter(void)
{
  printf("\nPress ENTER ...");
  while (getchar() != '\n')
    ;
}
//---------------------------------------------------------------------------
void usage(void)
{
  version();

	printf("Usage: %s [options] infile.ext\n", PROG_NAME);
  printf("\nCommon options:\n");
  printf("  -o outfile.ext  output file name\n");
  printf("  -append\t  append at the end of the output file\n");
  printf("  -m 1234\t  pack only with given modes [all modes are trying]\n");
  printf("  -lf firstLine\t  first microline of screen [0]\n");
  printf("  -ll lastLine\t  last microline of screen [255]\n");
  printf("  -a address\t  insert unpack routine at given address\n");
  printf("  -ptp\t\t  create PTP file\n");
  printf("  -vb\t\t  verbose\n");
  printf("  -h\t\t  usage information\n");
  printf("  -v\t\t  version info\n");
  printf("\nAll numeric parameters must be nonnegative integers and can be in three forms:\n");
  printf("  decimal (e.g. 12345)\n");
  printf("  octal (e.g. 01234)\n");
  printf("  hexadecimal (e.g. 0x1234)\n\n");

  return;
}
//---------------------------------------------------------------------------

