//---------------------------------------------------------------------------
#include "common.h"
//---------------------------------------------------------------------------
#define SCREEN_LINE_WIDTH     64
#define MICRO_LINE_WIDTH      48

#define IM1_DICT_SIZE         4096
#define IM1_PHRASE_MIN        3
#define IM1_PHRASE_MAX        (IM1_PHRASE_MIN + 7)        // 10
#define IM1_PHRASE_SHIFT      4
#define IM1_PHRASE_MASK       0x70
#define IM1_OFFSET_SHIFT      0
#define IM1_OFFSET_MASK       0x0F

#define IM2_DICT_SIZE         2048
#define IM2_PHRASE_MIN        3
#define IM2_PHRASE_MAX        (IM2_PHRASE_MIN + 15)       // 18
#define IM2_PHRASE_SHIFT      0
#define IM2_PHRASE_MASK       0x0F
#define IM2_OFFSET_SHIFT      4
#define IM2_OFFSET_MASK       0x70

#define SHRINK_MIN            3
#define SHRINK_MAX            MICRO_LINE_WIDTH

#define NOT_PACKED_MAX        MICRO_LINE_WIDTH

#define FLAG_IMPLOD           0x00    // 0LLLAAAA alebo 0AAALLLL
#define FLAG_SHRINK_LITTLE    0xC0    // 11LLLLLL
// #define FLAG_SHRINK_BIG       0x80    // 10000000
#define FLAG_NO_PACKED        0x80    // 10LLLLLL

//---------------------------------------------------------------------------
int packBackward(void);
int pbFindBetter(BYTE *shrinkByte, int *implOff);
int pbShrink(BYTE *shrinkByte);
int pbImplod(int *implOff);
int pbImplodFindPhrase(int *implOff, BYTE *searchIn, int maxP, int maxF);

int packForward(void);
int pfFindBetter(BYTE *shrinkByte, int *implOff);
int pfShrink(BYTE *shrinkByte);
int pfImplod(int *implOff);
int pfImplodFindPhrase(int *implOff, BYTE *searchIn, int maxP, int maxF);

//---------------------------------------------------------------------------
// globalne premenne pre "pack.c"
int dictSize, phraseMin, phraseMax, phraseShift, phraseMask, offsetShift, offsetMask;
BYTE *pFirst, *pLast, *pAct, *pTarget, *pEnd;

//---------------------------------------------------------------------------
int pack(BYTE *buffOut, BYTE *buffIn, int firstLine, int lastLine, int packMode)
{
  int outLen;

  switch (packMode) {
    case PACK_MODE_1 :
    case PACK_MODE_3 :
      dictSize = IM1_DICT_SIZE;
      phraseMin = IM1_PHRASE_MIN;
      phraseMax = IM1_PHRASE_MAX;
      phraseShift = IM1_PHRASE_SHIFT;
      phraseMask = IM1_PHRASE_MASK;
      offsetShift = IM1_OFFSET_SHIFT;
      offsetMask = IM1_OFFSET_MASK;
      break;

    case PACK_MODE_2 :
    case PACK_MODE_4 :
      dictSize = IM2_DICT_SIZE;
      phraseMin = IM2_PHRASE_MIN;
      phraseMax = IM2_PHRASE_MAX;
      phraseShift = IM2_PHRASE_SHIFT;
      phraseMask = IM2_PHRASE_MASK;
      offsetShift = IM2_OFFSET_SHIFT;
      offsetMask = IM2_OFFSET_MASK;
      break;

    default :
      return ERR_BAD_PACK_MODE;
  }

  pFirst = buffIn + firstLine * SCREEN_LINE_WIDTH;
  pLast = buffIn + lastLine * SCREEN_LINE_WIDTH + MICRO_LINE_WIDTH - 1;

  switch (packMode) {
    case PACK_MODE_1 :
    case PACK_MODE_2 :
      pTarget = buffOut + MAX_FILE_LENGTH - 1;
      pEnd = buffOut + 2;
      outLen = packBackward();
      if (outLen == ERR_OK) {
        pFirst = buffOut + 2;
        pLast = buffOut + MAX_FILE_LENGTH - 1;
        outLen = pLast - pTarget;
        *(WORD *)buffOut = (WORD)outLen;
        while (pTarget < pLast)     // presun spakovany blok na zaciatok buffra
          *pFirst++ = *++pTarget;
        outLen += 2;
      }
      break;

    case PACK_MODE_3 :
    case PACK_MODE_4 :
      pTarget = buffOut + 2;
      pEnd = buffOut + MAX_FILE_LENGTH - 1;
      outLen = packForward();
      if (outLen == ERR_OK) {
        outLen = pTarget - buffOut - 2;
        *(WORD *)buffOut = (WORD)outLen;
        outLen += 2;
      }
      break;
  }

  return outLen;
}
//---------------------------------------------------------------------------
// prevedie spakovanie bloku dat odzadu a vrati jeho novu dlzku.
// Ak sa pakovanie nepodari, vrati sa chyba ERR_INEFFECTIVE.
int packBackward(void)
{
  int noPack, implOffset, cnt;
  BYTE shrByte;

  pAct = pLast; // ideme odzadu
  noPack = 0;
  while (pAct >= pFirst) {
    cnt = pbFindBetter(&shrByte, &implOffset);
    if (cnt < 0) {      // nepakovatelne
      *pTarget-- = *pAct--;
      noPack++;
      if (noPack == NOT_PACKED_MAX) {
        if (pTarget < pEnd) // problem - data sa nedaju "rozumne" spakovat
          return ERR_INEFFECTIVE;
        *pTarget-- = (BYTE)(FLAG_NO_PACKED | noPack);
        if (pTarget < pEnd) // problem - data sa nedaju "rozumne" spakovat
          return ERR_INEFFECTIVE;
        noPack = 0;
      }
    }
    else {  // pakovatelne
      if (noPack > 0) { // ukonci pripadny nepakovatelny blok
        if (pTarget < pEnd) // problem - data sa nedaju "rozumne" spakovat
          return ERR_INEFFECTIVE;
        *pTarget-- = (BYTE)(FLAG_NO_PACKED | noPack);
        if (pTarget < pEnd) // problem - data sa nedaju "rozumne" spakovat
          return ERR_INEFFECTIVE;
        noPack = 0;
      }

      pAct -= cnt;  // posun aktualnej adresy
      if (implOffset >= 0) {  // Imploding
        *pTarget-- = (BYTE)implOffset;
        *pTarget-- = (BYTE)((((cnt - phraseMin) << phraseShift) & phraseMask)
                            | (((implOffset / 256) << offsetShift) & offsetMask));
      }
//      else if (cnt > SHRINK_MAX_LITTLE) { // dlhy Shrink
//        *pTarget-- = shrByte;
//        *pTarget-- = (BYTE)(cnt - SHRINK_MAX_LITTLE - 1);
//        *pTarget-- = (BYTE)FLAG_SHRINK_BIG;
//      }
      else {  // kratky Shrink
        *pTarget-- = shrByte;
        *pTarget-- = (BYTE)(FLAG_SHRINK_LITTLE | (cnt - SHRINK_MIN));
      }
    }

    if (((pAct - pFirst) & 0x3F) == 0x3F) {
      pAct -= (SCREEN_LINE_WIDTH - MICRO_LINE_WIDTH);
      if (noPack > 0) { // ukonci pripadny nepakovatelny blok
        *pTarget-- = (BYTE)(FLAG_NO_PACKED | noPack);
        noPack = 0;
      }
    }
  }

  if (noPack > 0) // ukonci pripadny nepakovatelny blok
    *pTarget-- = (BYTE)(FLAG_NO_PACKED | noPack);

  return ERR_OK;
}
//---------------------------------------------------------------------------
// Pokusi sa najst na aktualnej pozicii pakovatelnu sekvenciu. Pokial sa takato
// sekvencia nenajde, vrati -1. Pokial sa najde, vrati jej dlzku. Ak sa jednalo
// o Imploding, na adresu implOff je ulozeny offset, kde sa nachadza dana fraza.
// Ak sa jedna o Shrinking, na adresu implOff je ulozena hodnota -1 a na adresu
// shrinkByte je ulozeny opakujuci sa byte.
int pbFindBetter(BYTE *shrinkByte, int *implOff)
{
  int shrCnt, implCnt, implOffset;

  *implOff = -1;

  shrCnt = pbShrink(shrinkByte); // skus Shrink
  if (shrCnt >= phraseMax)       // ak je pocet vacsi ako maximalna dlzka frazy,
    return shrCnt;               // nerob Implod a vrat sa rovno s tymto poctom

  implCnt = pbImplod(&implOffset);  // skus Implod
  if (implCnt > shrCnt) {           // ak je impolod uspesnejsi,
    *implOff = implOffset;          // nastav offset ku najdenej fraze
    return implCnt;                 // vrat sa s dlzkou frazy
  }

  return shrCnt;  // inak sa vrat s hodnotou so Shrinku, to moze byt aj -1
}
//---------------------------------------------------------------------------
// Vyhlada na aktualnej pozicii minimalne SHRINK_MIN a maximalne SHRINK_MAX_BIG
// rovnakych bytov a vrati ich pocet. Opakujuci sa byte je ulozeny na adresu
// shrinkByte. Pokial sa rovnake byty nenajdu, vrati -1.
int pbShrink(BYTE *shrinkByte)
{
  int cnt = -1;
  BYTE val, *p, *pMin;

  pMin = pFirst + ((pAct - pFirst) & 0x3FC0) - 1;
  if ((pAct - pMin) >= SHRINK_MIN
      && *pAct == *(pAct - 1) && *pAct == *(pAct - 2)) { // SHRINK_MIN = 3
    cnt = SHRINK_MIN;
    val = *pAct;
    p = pAct - 3;
    while (p > pMin && cnt < SHRINK_MAX) {
      if (*p != val)
        break;
      p--;
      cnt++;
    }
    *shrinkByte = val;
  }

  return cnt;
}
//---------------------------------------------------------------------------
// Od aktualnej pozicie vyhlada frazu, ktora sa nachadza dalej k nizsim adresam.
// Najdena fraza musi mat minimalnu dlzku phraseMin a moze mat maximalne phraseMax.
// Ak je fraza najdena, na adresu implOff sa ulozi offset k cielovej fraze k nizsim
// adresam a vratena je dlzka tejto frazy. Pokial sa fraza nenajde, vrati sa -1.
int pbImplod(int *implOff)
{
  int cnt, cntI, maxP, maxF, implOffset;
  BYTE *pMin, *p;

  pMin = pFirst + ((pAct - pFirst) & 0x3FC0) - 1;
  maxP = pAct - pMin;
  maxF = maxP - 1;
  if (maxF < phraseMin)
    return -1;

  cnt = -1;
  p = pAct - 1;
  while (p > pFirst) {
    cntI = pbImplodFindPhrase(&implOffset, p, maxP, maxF);
    if (cntI > cnt) {
      cnt = cntI;
      *implOff = implOffset;
      if (cnt == phraseMax || cnt == maxP)
        break;
    }

    p = pFirst + ((p - pFirst) & 0x3FC0) - 1 - (SCREEN_LINE_WIDTH - MICRO_LINE_WIDTH);
    maxF = MICRO_LINE_WIDTH;
  }

  return cnt;
}
//---------------------------------------------------------------------------
int pbImplodFindPhrase(int *implOff, BYTE *searchIn, int maxP, int maxF)
{
  int cnt, cntD, implOffset, cntP, cntC;
  BYTE val, *p, *p1, *p2;

  val = *pAct;  // prvy byte frazy
  p = searchIn; // adresa kde hladat

  cnt = -1;
  cntD = maxF; // vypocitaj maximalnu velkost "okna" pre hladanie frazy

  while (TRUE) {
    while (--cntD > 0 && *p-- != val) // hladaj prvy zhodny byte frazy
      ;
    // ak je zostavajuca velkost potencialnej frazy kratsia ako minimalna alebo
    // ako najvacsia najdena, rovno to ukonci, nema to zmysel
    if (cntD <= cnt || cntD < phraseMin)
      break;

    implOffset = pAct - p - 1;  // vypocitaj offset k najdenemu zhodnemu bytu
    if (implOffset >= dictSize) // ak offset prekracuje velkost "okna",
      break;                    // ukonci to

    p1 = p;                     // hladaj dalsie rovnake byty
    p2 = pAct - 1;
    if (maxP < phraseMax)  // maximalne vsak do maxima frazy alebo zostatku
      cntC = maxP - 1;
    else
      cntC = phraseMax - 1;
    if (cntD < cntC)
      cntC = cntD;
    cntP = cntC + 1;
    while (cntC > 0) {          // hladaj frazu
      if (*p1 != *p2)
        break;
      p1--;
      p2--;
      cntC--;
    }
    cntC = cntP - cntC;                     // dlzka najdenej frazy
    if (cntC >= phraseMin && cntC > cnt) {  // ak je najdena fraza dostatocne
      cnt = cntC;                           // velka, akceptuj ju a zapamataj
      *implOff = implOffset;
    }

    if (cnt == cntP)  // ak si na konci "okna" a vacsia fraza uz nebude,
      break;          // ukonci to
  }

  return cnt;
}
//---------------------------------------------------------------------------
//***************************************************************************
//---------------------------------------------------------------------------
// Prevedie spakovanie bloku dat spredu a vrati jeho novu dlzku.
// Ak sa pakovanie nepodari, vrati sa chyba ERR_INEFFECTIVE.
int packForward(void)
{
  int noPack, implOffset, cnt;
  BYTE shrByte;

  pAct = pFirst; // ideme spredu
  noPack = 0;
  while (pAct <= pLast) {
    cnt = pfFindBetter(&shrByte, &implOffset);
    if (cnt < 0) {      // nepakovatelne
      *pTarget++ = *pAct++;
      noPack++;
      if (noPack == NOT_PACKED_MAX) {
        if (pTarget > pEnd) // problem - data sa nedaju "rozumne" spakovat
          return ERR_INEFFECTIVE;
        *pTarget++ = (BYTE)(FLAG_NO_PACKED | noPack);
        if (pTarget > pEnd) // problem - data sa nedaju "rozumne" spakovat
          return ERR_INEFFECTIVE;
        noPack = 0;
      }
    }
    else {  // pakovatelne
      if (noPack > 0) { // ukonci pripadny nepakovatelny blok
        if (pTarget > pEnd) // problem - data sa nedaju "rozumne" spakovat
          return ERR_INEFFECTIVE;
        *pTarget++ = (BYTE)(FLAG_NO_PACKED | noPack);
        if (pTarget > pEnd) // problem - data sa nedaju "rozumne" spakovat
          return ERR_INEFFECTIVE;
        noPack = 0;
      }

      pAct += cnt;  // posun aktualnej adresy
      if (implOffset >= 0) {  // Imploding
        *pTarget++ = (BYTE)implOffset;
        *pTarget++ = (BYTE)((((cnt - phraseMin) << phraseShift) & phraseMask)
                            | (((implOffset / 256) << offsetShift) & offsetMask));
      }
//      else if (cnt > SHRINK_MAX_LITTLE) { // dlhy Shrink
//        *pTarget++ = shrByte;
//        *pTarget++ = (BYTE)(cnt - SHRINK_MAX_LITTLE - 1);
//        *pTarget++ = (BYTE)FLAG_SHRINK_BIG;
//      }
      else {  // kratky Shrink
        *pTarget++ = shrByte;
        *pTarget++ = (BYTE)(FLAG_SHRINK_LITTLE | (cnt - SHRINK_MIN));
      }
    }

    if (((pAct - pFirst) & 0x3F) == 0x30) {
      pAct += (SCREEN_LINE_WIDTH - MICRO_LINE_WIDTH);
      if (noPack > 0) { // ukonci pripadny nepakovatelny blok
        *pTarget++ = (BYTE)(FLAG_NO_PACKED | noPack);
        noPack = 0;
      }
    }
  }

  if (noPack > 0) // ukonci pripadny nepakovatelny blok
    *pTarget++ = (BYTE)(FLAG_NO_PACKED | noPack);

  return ERR_OK;
}
//---------------------------------------------------------------------------
// Pokusi sa najst na aktualnej pozicii pakovatelnu sekvenciu. Pokial sa takato
// sekvencia nenajde, vrati -1. Pokial sa najde, vrati jej dlzku. Ak sa jednalo
// o Imploding, na adresu implOff je ulozeny offset, kde sa nachadza dana fraza.
// Ak sa jedna o Shrinking, na adresu implOff je ulozena hodnota -1 a na adresu
// shrinkByte je ulozeny opakujuci sa byte.
int pfFindBetter(BYTE *shrinkByte, int *implOff)
{
  int shrCnt, implCnt, implOffset;

  *implOff = -1;

  shrCnt = pfShrink(shrinkByte); // skus Shrink
  if (shrCnt >= phraseMax)       // ak je pocet vacsi ako maximalna dlzka frazy,
    return shrCnt;               // nerob Implod a vrat sa rovno s tymto poctom

  implCnt = pfImplod(&implOffset);  // skus Implod
  if (implCnt > shrCnt) {           // ak je impolod uspesnejsi,
    *implOff = implOffset;          // nastav offset ku najdenej fraze
    return implCnt;                 // vrat sa s dlzkou frazy
  }

  return shrCnt;  // inak sa vrat s hodnotou so Shrinku, to moze byt aj -1
}
//---------------------------------------------------------------------------
// Vyhlada na aktualnej pozicii minimalne SHRINK_MIN a maximalne SHRINK_MAX
// rovnakych bytov a vrati ich pocet. Opakovany byte je ulozeny na adresu shrinkByte.
// Pokial sa rovnake byty nenajdu, vrati -1.
int pfShrink(BYTE *shrinkByte)
{
  int cnt = -1;
  BYTE val, *p, *pMax;

  pMax = pFirst + ((pAct - pFirst) & 0x3FC0) + MICRO_LINE_WIDTH;
  if ((pMax - pAct) >= SHRINK_MIN
      && *pAct == *(pAct + 1) && *pAct == *(pAct + 2)) { // SHRINK_MIN = 3
    cnt = SHRINK_MIN;
    val = *pAct;
    p = pAct + 3;
    while (p < pMax && cnt < SHRINK_MAX) {
      if (*p != val)
        break;
      p++;
      cnt++;
    }
    *shrinkByte = val;
  }

  return cnt;
}
//---------------------------------------------------------------------------
// Od aktualnej pozicie vyhlada frazu, ktora sa nachadza dalej k vyssim adresam.
// Najdena fraza musi mat minimalnu dlzku phraseMin a moze mat maximalne phraseMax.
// Ak je fraza najdena, na adresu implOff sa ulozi offset k cielovej fraze k vyssim
// adresam a vratena je dlzka tejto frazy. Pokial sa fraza nenajde, vrati sa -1.
int pfImplod(int *implOff)
{
  int cnt, cntI, maxP, maxF, implOffset;
  BYTE *pMax, *p;

  pMax = pFirst + ((pAct - pFirst) & 0x3FC0) + MICRO_LINE_WIDTH;
  maxP = pMax - pAct;
  maxF = maxP - 1;
  if (maxF < phraseMin)
    return -1;

  cnt = -1;
  p = pAct + 1;
  while (p < pLast) {
    cntI = pfImplodFindPhrase(&implOffset, p, maxP, maxF);
    if (cntI > cnt) {
      cnt = cntI;
      *implOff = implOffset;
      if (cnt == phraseMax || cnt == maxP)
        break;
    }

    p = pFirst + ((p - pFirst) & 0x3FC0) + SCREEN_LINE_WIDTH;
    maxF = MICRO_LINE_WIDTH;
  }

  return cnt;
}
//---------------------------------------------------------------------------
int pfImplodFindPhrase(int *implOff, BYTE *searchIn, int maxP, int maxF)
{
  int cnt, cntD, implOffset, cntP, cntC;
  BYTE val, *p, *p1, *p2;

  val = *pAct;  // prvy byte frazy
  p = searchIn; // adresa kde hladat

  cnt = -1;
  cntD = maxF;
  while (TRUE) {
    while (--cntD > 0 && *p++ != val) // hladaj prvy zhodny byte frazy
      ;
    // ak je zostavajuca velkost potencialnej frazy kratsia ako minimalna alebo
    // ako najvacsia najdena, rovno to ukonci, nema to zmysel
    if (cntD <= cnt || cntD < phraseMin)
      break;

    implOffset = p - pAct - 1;  // vypocitaj offset k najdenemu zhodnemu bytu
    if (implOffset >= dictSize) // ak offset prekracuje velkost "okna",
      break;                    // ukonci to

    p1 = p;                     // hladaj dalsie rovnake byty
    p2 = pAct + 1;
    if (maxP < phraseMax)  // maximalne vsak do maxima frazy alebo zostatku
      cntC = maxP - 1;
    else
      cntC = phraseMax - 1;
    if (cntD < cntC)
      cntC = cntD;
    cntP = cntC + 1;
    while (cntC > 0) {          // hladaj frazu
      if (*p1 != *p2)
        break;
      p1++;
      p2++;
      cntC--;
    }
    cntC = cntP - cntC;                     // dlzka najdenej frazy
    if (cntC >= phraseMin && cntC > cnt) {  // ak je najdena fraza dostatocne
      cnt = cntC;                           // velka, akceptuj ju a zapamataj
      *implOff = implOffset;
    }

    if (cnt == cntP)  // ak si na konci "okna" a vacsia fraza uz nebude,
      break;          // ukonci to
  }

  return cnt;
}
//---------------------------------------------------------------------------

